from snovault import collection, load_schema  # , calculated_property
from .base import Item


# XXX: These constants could be defined on CompoundSearchBuilder, but thought they may be
#      more useful here/more logically belong. - Will
FLAGS = 'flags'
FILTER_BLOCKS = 'filter_blocks'


@collection(
    name='filter-sets',
    unique_key='filter_set:title',
    properties={
        'title': 'Filter Sets',
        'description': 'Filter Set for combining multiple queries'
    }
)
class FilterSet(Item):
    """The class to store information about 4DN file formats"""
    item_type = 'filter_set'
    schema = load_schema('snovault:schemas/filter_set.json')
    embedded_list = []

    class Collection(Item.Collection):
        pass



from collections import OrderedDict
from itertools import chain
from urllib.parse import urlparse

from pyramid.httpexceptions import HTTPNotFound
from pyramid.view import view_config

from .etag import etag_app_version_effective_principals
from .interfaces import (
    COLLECTIONS,
    TYPES,
)
from .util import debug_log


def includeme(config):
    config.add_route('schemas', '/profiles/')
    config.add_route('schema', '/profiles/{type_name}.json')
    config.scan(__name__)


def _annotated_schema(type_info, request):
    """
    Add some extra annotiation to a schema obtained through given TypeInfo.
    Specifically, adds links to the /terms/ page and inheritance information
    through the `children` and `rdfs:subClassOf` properties. Also flags fields
    by write permission with the `readonly` field, if applicable.
    TODO: add flagging for user restricted fields once role-based field viewing
    is implemented.

    Args:
        type_info (TypeInfo): for an item type. See snovault.type_info.py
        request (Request): the current Request

    Returns:
        dict: the annotated schema
    """
    schema = type_info.schema.copy()
    schema['@type'] = ['JSONSchema']
    jsonld_base = request.registry.settings['snovault.jsonld.terms_namespace']
    schema['rdfs:seeAlso'] = urlparse(jsonld_base).path + type_info.name
    # add links to profiles of children schemas
    schema['children'] = [
        '/profiles/' + t_name + '.json' for t_name in type_info.child_types
    ]

    if type_info.factory is None:
        return schema

    # use first base_type that is not this type itself to handle abstract
    found_subtype = None
    for subtype in type_info.base_types:
        if subtype != type_info.name:
            found_subtype = subtype
            break
    if found_subtype:
        schema['rdfs:subClassOf'] = '/profiles/' + found_subtype + '.json'
    # add abstract flag to know if the profile represents abstract item
    schema['isAbstract'] = type_info.is_abstract

    collection = request.registry[COLLECTIONS][type_info.name]
    properties = OrderedDict()
    # add a 'readonly' flag to fields that the current user cannot write
    for k, v in schema['properties'].items():
        if 'permission' in v:
            if not request.has_permission(v['permission'], collection):
                v = v.copy()
                v['readonly'] = True
        properties[k] = v
    schema['properties'] = properties
    return schema


@view_config(route_name='schema', request_method='GET',
             decorator=etag_app_version_effective_principals)
@debug_log
def schema(context, request):
    """
    /profiles/{type_name}.json -- view for the profile of a specific item type
    A bit inefficient, but need to use the TypeInfo (not AbstractTypeInfo)
    to get the correct schema. To do this, iterate through all registered
    types until we find the one with matching item_type (given by type_name).
    This allows this endpoint to work with item name (e.g. MyItem) or item_type
    (e.g. my_item)
    """
    type_name = request.matchdict['type_name']
    types = request.registry[TYPES]
    found_type_info = None
    all_item_types = chain(types.by_item_type.values(),
                           types.by_abstract_type.values())
    for type_info in all_item_types:
        # handle both item name and item type inputs to the route (both valid)
        if type_info.name == type_name or type_info.item_type == type_name:
            found_type_info = type_info
            break
    if found_type_info is None:
        raise HTTPNotFound(type_name)
    return _annotated_schema(type_info, request)



@view_config(route_name='schemas', request_method='GET',
             decorator=etag_app_version_effective_principals)
@debug_log
def schemas(context, request):
    """
    /profiles/ view for viewing all schemas. Leverages the TypeInfo objects
    for regular classes using registry[TYPES].by_item_type and for abstract
    classes by using registry[TYPES].by_abstract_type
    """
    types = request.registry[TYPES]
    schemas = {}
    all_item_types = chain(types.by_item_type.values(),
                           types.by_abstract_type.values())
    for type_info in all_item_types:
        name = type_info.name
        schemas[name] = _annotated_schema(type_info, request)
    return schemas



import argparse
import logging
import structlog

from dcicutils.env_utils import permit_load_data
from dcicutils.common import APP_CGAP

from pyramid.paster import get_app
from pyramid.path import DottedNameResolver
from .. import configure_dbsession


log = structlog.getLogger(__name__)


EPILOG = __doc__


# should be overridden in downstream application to pass a different app
def load_data_should_proceed(env, allow_prod, app=None):
    """ Returns True on whether or not load_data should proceed.

    :param env: env we are on
    :param allow_prod: prod argument from argparse, defaults to False
    :param app: app type, one of cgap, fourfront (enums from dcicutils.common)
    :return: True if load_data should continue, False otherwise
    """
    if not app:
        app = APP_CGAP  # this fallback is somewhat reasonable
    return permit_load_data(envname=env, allow_prod=allow_prod, orchestrated_app=app)  # noqa


def main(simulated_args=None):
    logging.basicConfig()
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('encoded').setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Load Data", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('config_uri', help="path to configfile")
    parser.add_argument('--prod', action='store_true',
                        help="must be set to confirm this action is intended to happen on a production server")
    parser.add_argument('--overwrite', action='store_true',
                        help="must be set to update existing uuids with patch")
    args = parser.parse_args(simulated_args)

    # get the pyramids app
    app = get_app(args.config_uri, args.app_name)

    # create db schema
    configure_dbsession(app)

    env = app.registry.settings.get('env.name', '')

    load_test_data = app.registry.settings.get('load_test_data')
    allow_prod = args.prod
    log.info("load_data: load_test_data function is %s" % (load_test_data))
    load_test_data = DottedNameResolver().resolve(load_test_data)

    if load_data_should_proceed(env, allow_prod):
        load_test_data(app, args.overwrite)


if __name__ == "__main__":
    main()



# Authentication related functions which may be overriden by an implementing app,
# e.g. Foursight or CGAP portal, using the dcicutils project_utils mechanism.

from ..authentication import login, logout

class SnovaultProjectAuthentication:

    def login(self, context, request, *, samesite):
        return login(context, request, samesite=samesite)

    def logout(self, context, request):
        return logout(context, request)

    def namespaced_authentication_policy_authenticated_userid(self, namespaced_authentication_policy, request, set_user_info_property):
        return namespaced_authentication_policy._authenticated_userid_implementation(request, set_user_info_property)

    def namespaced_authentication_policy_unauthenticated_userid(self, namespaced_authentication_policy, request):
        return namespaced_authentication_policy._unauthenticated_userid_implementation(request)

    def note_auth0_authentication_policy_unauthenticated_userid(self, auth0_authentication_policy, request, email, id_token):
        pass

    # TODO: Maybe something like ...
    # def __init__(self):
    #    self.login_policy = 


# TODO: Maybe something like ...
# def SnovaultNamespacedAuthenticationPolicy:
#     def __init__(self, app_project):
#         self.app_project = app_project
#     def authenticated_userid(self, namespaced_authentication_policy, request, set_user_info_property):
#         return namespaced_authentication_policy._authenticated_userid_implementation(request, set_user_info_property)



import os.path
import sys
import subprocess

import atexit
import shutil
import tempfile

from urllib.parse import urlencode


def initdb(datadir, prefix='', echo=False):
    init_args = [
        os.path.join(prefix, 'initdb'),
        '-D', datadir,
        '-U', 'postgres',
        '--auth=trust',
    ]
    output = subprocess.check_output(
        init_args,
        close_fds=True,
        stderr=subprocess.STDOUT,
    )
    if echo:
        print(output.decode('utf-8'))


DEFAULT_SNOVAULT_DB_TEST_PORT = 5440

_custom_port = os.environ.get("SNOVAULT_DB_TEST_PORT", None)
try:
    SNOVAULT_DB_TEST_PORT = int(_custom_port) if _custom_port else DEFAULT_SNOVAULT_DB_TEST_PORT
except (ValueError, TypeError):
    print("Bad SNOVAULT_DB_TEST_PORT=%r. Using %s." % (_custom_port, DEFAULT_SNOVAULT_DB_TEST_PORT))
    SNOVAULT_DB_TEST_PORT = DEFAULT_SNOVAULT_DB_TEST_PORT


# These have some wired behavior in the server_process function below. It may not work
# to override them, since the 'postgres' command doesn't take a username parameter.
# Permissions and other issues might come into play, since the db is initialized for
# a postgres/postgres username/dbname, and to use another db might require additional
# initialization. -kmp 14-Mar-2020

SNOVAULT_DB_TEST_HOSTNAME = ''
SNOVAULT_DB_TEST_USERNAME = 'postgres'
SNOVAULT_DB_TEST_DBNAME = 'postgres'
SNOVAULT_DB_TEST_DATADIR = None


def make_snovault_db_test_url(username=SNOVAULT_DB_TEST_USERNAME,
                              port=SNOVAULT_DB_TEST_PORT,
                              hostname=SNOVAULT_DB_TEST_HOSTNAME,
                              dbname=SNOVAULT_DB_TEST_DBNAME,
                              datadir=SNOVAULT_DB_TEST_DATADIR):
    query_string = "?" + urlencode({"host": datadir}) if datadir else ""
    return "postgresql://%s@%s:%s/%s%s" % (username, hostname, port, dbname, query_string)


def server_process(datadir, prefix='', echo=False):

    postgres_command = os.path.join(prefix, 'postgres')

    command = [
        postgres_command,
        '-D', datadir,
        '-F',  # no fsync
        '-h', SNOVAULT_DB_TEST_HOSTNAME,
        '-k', datadir,
        '-p', str(SNOVAULT_DB_TEST_PORT),
    ]
    process = subprocess.Popen(
        command,
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    success_line = b'database system is ready to accept connections\n'

    lines = []
    for line in iter(process.stdout.readline, b''):
        if echo:
            sys.stdout.write(line.decode('utf-8'))
        lines.append(line)
        if line.endswith(success_line):
            break
    else:
        code = process.wait()
        msg = ('Process return code: %d\n' % code) + b''.join(lines).decode('utf-8')
        raise Exception(msg)

    if not echo:
        process.stdout.close()

    if echo:
        print('Created: %s' % make_snovault_db_test_url(datadir=datadir))

    return process


def main():
    datadir = tempfile.mkdtemp()

    def clean_datadir():
        shutil.rmtree(datadir)
        print('Cleaned dir: %s' % datadir)

    print('Starting in dir: %s' % datadir)
    try:
        process = server_process(datadir, echo=True)
    except BaseException:
        # If there's an error setting up the server process, rush an early cleanup.
        clean_datadir()
        raise

    @atexit.register
    def cleanup_process():
        try:
            if process.poll() is None:
                process.terminate()
                for output_line in process.stdout:
                    sys.stdout.write(output_line.decode('utf-8'))
                process.wait()
        finally:
            clean_datadir()

    print('Started. ^C to exit.')

    try:
        for line in iter(process.stdout.readline, b''):
            sys.stdout.write(line.decode('utf-8'))
    except KeyboardInterrupt:
        raise SystemExit(0)


if __name__ == '__main__':
    main()



import logging
import pytest

from ..project_defs import C4ProjectRegistry
from ..elasticsearch.indexer_queue import QueueManager


# required so that db transactions are properly rolled back in tests
@pytest.fixture(autouse=True)
def autouse_external_tx(external_tx):
    pass


# def _check_server_is_up(output):
#     """ Polls the given output file to detect
#
#         :args output: file to which server is piping out
#         :returns: True if server is up, False if failed
#     """
#     tries = 5
#     while tries > 0:
#         output.seek(0)  # should be first thing to be output.
#         out = output.read()
#         if 'Running' in out.decode('utf-8'):
#             return True
#         tries -= 1
#         time.sleep(1)  # give it a sec
#     return False
#
#
# @pytest.fixture(scope='session', autouse=True)
# def start_moto_server_sqs():
#     """
#     Spins off a moto server running sqs, yields to the tests and cleans up.
#     """
#     delete_sqs_url = 'SQS_URL' not in os.environ
#     old_sqs_url = os.environ.get('SQS_URL', None)
#     server_output = tempfile.TemporaryFile()
#     server = None
#     try:
#         try:
#             os.environ['SQS_URL'] = 'http://localhost:3000'  # must exists globally because of MPIndexer
#             server_args = ['poetry', 'run', 'moto_server', 'sqs', '-p3000']
#             server = subprocess.Popen(server_args, stdout=server_output, stderr=server_output)
#             time.sleep(5)
#             assert _check_server_is_up(server_output)
#         except AssertionError:
#             server_output.seek(0)
#             raise AssertionError(server_output.read().decode('utf-8'))
#         except Exception as e:
#             raise Exception('Encountered an exception bringing up the server: %s' % str(e))
#
#         yield  # run tests
#
#     finally:
#         if delete_sqs_url:
#             del os.environ['SQS_URL']
#         else:
#             os.environ['SQS_URL'] = old_sqs_url
#         if server:
#             server.terminate()


def pytest_configure():
    logging.basicConfig()
    logging.getLogger('snovault').setLevel(logging.INFO)
    QueueManager.PURGE_QUEUE_SLEEP_FOR_SAFETY = True



import random
import uuid

from dcicutils.misc_utils import utc_now_str
from jsonschema_serialize_fork import NO_DEFAULT
from pyramid.path import DottedNameResolver
from pyramid.threadlocal import get_current_request
from snovault.schema_utils import server_default
from snovault import COLLECTIONS  # , ROOT
from string import digits  # , ascii_uppercase
from .project_app import app_project


ACCESSION_FACTORY = __name__ + ':accession_factory'
ACCESSION_PREFIX = app_project().ACCESSION_PREFIX
ACCESSION_TEST_PREFIX = 'TST'


def includeme(config):
    accession_factory = config.registry.settings.get('accession_factory')
    if accession_factory:
        factory = DottedNameResolver().resolve(accession_factory)
    else:
        factory = enc_accession
    config.registry[ACCESSION_FACTORY] = factory


# XXX: This stuff is all added based on the serverDefault identifier in the schemas
# removing it altogether will totally break our code


@server_default
def userid(instance, subschema):  # args required by jsonschema-serialize-fork
    return _userid()


def _userid():
    request = get_current_request()
    for principal in request.effective_principals:
        if principal.startswith('userid.'):
            return principal[7:]
    return NO_DEFAULT


@server_default
def now(instance, subschema):  # args required by jsonschema-serialize-fork
    return utc_now_str()


@server_default
def uuid4(instance, subschema):
    return str(uuid.uuid4())


@server_default
def accession(instance, subschema):
    if 'external_accession' in instance:
        return NO_DEFAULT
    request = get_current_request()
    factory = request.registry[ACCESSION_FACTORY]
    # With 17 576 000 options
    ATTEMPTS = 10
    for attempt in range(ATTEMPTS):
        new_accession = factory(subschema['accessionType'])
        if new_accession in request.root:
            continue
        return new_accession
    raise AssertionError("Free accession not found in %d attempts" % ATTEMPTS)


def get_userid():
    """ Wrapper for the server_default 'userid' above so it is not called through SERVER_DEFAULTS in our code """
    return _userid()


def get_user_resource():
    request = get_current_request()
    userid_found = _userid()
    if userid_found == NO_DEFAULT:
        return NO_DEFAULT
    return request.registry[COLLECTIONS]['user'][userid_found]


def get_now():
    """ Wrapper for the server_default 'now' above so it is not called through SERVER_DEFAULTS in our code """
    return utc_now_str()


def add_last_modified(properties, **kwargs):
    """
        Uses the above two functions to add the last_modified information to the item
        May have no effect
        Allow someone to override the request userid (none in this case) by passing in a different uuid
        CONSIDER: `last_modified` (and `last_text_edited`) are not really 'server defaults' but rather system-managed fields.
    """

    userid = kwargs.get("userid", None)
    field_name_portion = kwargs.get("field_name_portion", "modified")

    last_field_name = "last_" + field_name_portion  # => last_modified
    by_field_name = field_name_portion + "_by"      # => modified_by
    date_field_name = "date_" + field_name_portion  # => date_modified

    try:
        last_modified = {
            by_field_name: get_userid(),
            date_field_name: get_now(),
        }
    except AttributeError:  # no request in scope ie: we are outside the core application.
        if userid:
            last_modified = {
                by_field_name: userid,
                date_field_name: get_now(),
            }
            properties[last_field_name] = last_modified
    else:
        # get_userid returns NO_DEFAULT if no userid
        if last_modified[by_field_name] != NO_DEFAULT:
            properties[last_field_name] = last_modified


#FDN_ACCESSION_FORMAT = (digits, digits, digits, ascii_uppercase, ascii_uppercase, ascii_uppercase)
FDN_ACCESSION_FORMAT = ['ABCDEFGHIJKLMNOPQRSTUVWXYZ123456789']*7

def enc_accession(accession_type):
    random_part = ''.join(random.choice(s) for s in FDN_ACCESSION_FORMAT)
    return ACCESSION_PREFIX + accession_type + random_part


TEST_ACCESSION_FORMAT = (digits, ) * 7


def test_accession(accession_type):
    """ Test accessions are generated on test.encodedcc.org
    """
    random_part = ''.join(random.choice(s) for s in TEST_ACCESSION_FORMAT)
    return 'TST' + accession_type + random_part



import pytest

from dcicutils.qa_utils import notice_pytest_fixtures


# Test for storage.keys

items = [
    {'name': 'one', 'accession': 'TEST1'},
    {'name': 'two', 'accession': 'TEST2'},
]

bad_items = [
    {'name': 'one', 'accession': 'BAD1'},
    {'name': 'bad', 'accession': 'TEST1'},
]


@pytest.fixture
def content(testapp):
    notice_pytest_fixtures(testapp)
    url = '/testing-keys/'
    for item in items:
        testapp.post_json(url, item, status=201)


@pytest.mark.parametrize('item', items)
def test_unique_key(testapp, content, item):
    notice_pytest_fixtures(testapp, content)
    url = '/testing-keys/' + item['accession']
    res = testapp.get(url).maybe_follow()
    assert res.json['name'] == item['name']


def test_keys_bad_items(testapp, content):
    notice_pytest_fixtures(testapp, content)
    url = '/testing-keys/'
    for item in bad_items:
        testapp.post_json(url, item, status=409)


def test_keys_update(testapp):
    notice_pytest_fixtures(testapp)
    url = '/testing-keys/'
    item = items[0]
    res = testapp.post_json(url, item, status=201)
    location = res.location
    new_item = {'name': 'new_one', 'accession': 'NEW1'}
    testapp.put_json(location, new_item, status=200)
    testapp.post_json(url, item, status=201)
    testapp.put_json(location, item, status=409)


def test_keys_conflict(testapp):
    notice_pytest_fixtures(testapp)
    url = '/testing-keys/'
    item = items[1]
    initial = testapp.get(url).json['@graph']
    testapp.post_json(url, item, status=201)
    posted = testapp.get(url).json['@graph']
    assert len(initial)+1 == len(posted)
    conflict = testapp.post_json(url, item, status=409)
    assert conflict.status_code == 409
    conflicted = testapp.get(url).json['@graph']
    assert len(posted) == len(conflicted)



import base64
import codecs
import hashlib
import json
import os
import psycopg2
import psycopg2.extensions
import subprocess
import zope.sqlalchemy

from dcicutils.misc_utils import ignored, ignorable
from pyramid.config import Configurator
from pyramid.path import AssetResolver, caller_package
from pyramid.session import SignedCookieSessionFactory
from pyramid.settings import asbool
from pyramid_localroles import LocalRolesAuthorizationPolicy
from sqlalchemy import engine_from_config, event, orm  # , text as psql_text
from webob.cookies import JSONSerializer

from .interfaces import DBSESSION

from .elasticsearch import APP_FACTORY
from .json_renderer import json_renderer
from .storage import Base


STATIC_MAX_AGE = 0


def json_asset(spec, **kw):
    utf8 = codecs.getreader("utf-8")
    asset = AssetResolver(caller_package()).resolve(spec)
    return json.load(utf8(asset.stream()), **kw)


def changelogs(config):
    config.add_static_view(
        'profiles/changelogs', 'schemas/changelogs', cache_max_age=STATIC_MAX_AGE)


def configure_engine(settings):
    engine_url = settings['sqlalchemy.url']
    engine_opts = {}
    if engine_url.startswith('postgresql'):
        if settings.get('indexer_worker'):
            application_name = 'indexer_worker'
        elif settings.get('indexer'):
            application_name = 'indexer'
        else:
            application_name = 'app'
        engine_opts = dict(
            isolation_level='REPEATABLE READ',
            json_serializer=json_renderer.dumps,
            connect_args={'application_name': application_name}
        )
    engine = engine_from_config(settings, 'sqlalchemy.', **engine_opts)
    if engine.url.drivername == 'postgresql':
        timeout = settings.get('postgresql.statement_timeout')
        if timeout:
            timeout = int(timeout) * 1000
            set_postgresql_statement_timeout(engine, timeout)
    return engine


def set_postgresql_statement_timeout(engine, timeout: int = 20 * 1000):
    """
    Prevent Postgres waiting indefinitely for a lock.

    :param engine: a database engine
    :param timeout: a number of milliseconds to set for as statement_timeout
    """

    @event.listens_for(engine, 'connect')
    def connect(dbapi_connection, connection_record):
        ignored(connection_record)
        timeout_ms = timeout
        if not isinstance(timeout_ms, int):
            # This coercion will truncate 3.5 to 3, but so would the %d below,
            # and we have long used that. But the real purpose of introducing
            # this coercion is to get a ValueError if a string other than a
            # representation of a number slips through, to seal out accidental injection.
            # -kmp 6-Apr-2023
            timeout_ms = int(timeout_ms)
        cursor = dbapi_connection.cursor()
        try:
            # cursor: psycopg2.extensions.cursor
            # This call to psycopg2.extensions.cursor.execute expects a real string. Giving it an sqlalchemy.text
            # object will fail because something will try to do a boolean test, probably "if thing_to_execute:..."
            # and __bool__ is not defined on sqlalchemy.txt
            # Bottom line: Cannot wrap this string with psql_text(...) like we do elsewhere. It's not ready.
            # Might be we could do such a wrapper if we called execute on some other object.
            cursor.execute("SET statement_timeout = %d;" % timeout_ms)
        except psycopg2.Error:
            dbapi_connection.rollback()
        finally:
            cursor.close()
            dbapi_connection.commit()


def json_from_path(path, default=None):
    if path is None:
        return default
    return json.load(open(path))


def configure_dbsession(config, clear_data=False):
    """
    Create a sqlalchemy engine and a session that uses it, the latter of which
    is added to the registry. Handle some extra registration
    """
    settings = config.registry.settings
    DBSession = settings.pop(DBSESSION, None)

    # handle creating the database engine separately with indexer_worker
    if DBSession is None and not settings.get('indexer_worker'):
        engine = configure_engine(settings)

        # useful for test instances where we want to clear the data
        if clear_data:
            Base.metadata.drop_all(engine)

        if asbool(settings.get('create_tables', False)):
            Base.metadata.create_all(engine)

        DBSession = orm.scoped_session(orm.sessionmaker(bind=engine))
        zope.sqlalchemy.register(DBSession)

    config.registry[DBSESSION] = DBSession


def session(config):
    """ To create a session secret on the server:

    $ cat /dev/urandom | head -c 256 | base64 > session-secret.b64
    """
    settings = config.registry.settings
    if 'session.secret' in settings:
        secret = settings['session.secret'].strip()
        if secret.startswith('/'):
            secret = open(secret).read()
            secret = base64.b64decode(secret)
    else:
        secret = os.urandom(256)
    # auth_tkt has no timeout set
    # cookie will still expire at browser close
    if 'session.timeout' in settings:
        timeout = int(settings['session.timeout'])
    else:
        timeout = 60 * 60 * 24
    session_factory = SignedCookieSessionFactory(
        secret=secret,
        timeout=timeout,
        reissue_time=2**32,  # None does not work
        serializer=JSONSerializer(),
    )
    config.set_session_factory(session_factory)


def app_version(config):
    try:
        # For each of the next two subprocess calls, if there's an error, 'git' call will write to stderr,
        # but we don't care about that output, so we have muffled it by directing it to /dev/null.
        # If you're debugging this and want to know what's being ignored, comment out the 'stderr=' line.
        # -kmp 8-Feb-2020
        version = subprocess.check_output(
            ['git', '-C', os.path.dirname(__file__), 'describe'],
            stderr=subprocess.DEVNULL,
        ).decode('utf-8').strip()
        diff = subprocess.check_output(
            ['git', '-C', os.path.dirname(__file__), 'diff', '--no-ext-diff'],
            stderr=subprocess.DEVNULL,
        )
        if diff:
            version += '-patch' + hashlib.sha1(diff).hexdigest()[:7]
    except Exception:
        version = os.environ.get("ENCODED_VERSION", "test_version")
    config.registry.settings['snovault.app_version'] = version


def main(global_config, **local_config):
    """
    This function returns a Pyramid WSGI application.
    """
    settings = global_config
    settings.update(local_config)
    settings['snovault.jsonld.namespaces'] = json_asset('encoded:schemas/namespaces.json')
    settings['snovault.jsonld.terms_namespace'] = 'https://www.encodeproject.org/terms/'
    settings['snovault.jsonld.terms_prefix'] = 'encode'
    hostname_command = settings.get('hostname_command', '').strip()
    if hostname_command:
        hostname = subprocess.check_output(hostname_command, shell=True).strip()
        settings.setdefault('persona.audiences', '')
        settings['persona.audiences'] += '\nhttp://%s' % hostname
        settings['persona.audiences'] += '\nhttp://%s:6543' % hostname

    config = Configurator(settings=settings)
    config.registry[APP_FACTORY] = main  # used by mp_indexer
    config.include(app_version)

    config.include('pyramid_multiauth')  # must be before calling set_authorization_policy
    # Override default authz policy set by pyramid_multiauth
    config.set_authorization_policy(LocalRolesAuthorizationPolicy())
    config.include(session)
    config.include('.persona')

    config.include(configure_dbsession)
    config.include('snovault')
    config.commit()  # commit so search can override listing

    config.include('.renderers')

    if 'elasticsearch.server' in config.registry.settings:
        config.include('snovault.elasticsearch')

    config.include(changelogs)

    if asbool(settings.get('testing', False)):
        config.include('.tests.testing_views')

    # Load upgrades last so that all views (including testing views) are
    # registered.
    config.include('.upgrade')

    app = config.make_wsgi_app()

    workbook_filename = settings.get('load_workbook', '')
    load_test_only = asbool(settings.get('load_test_only', False))
    docsdir = settings.get('load_docsdir', None)
    if docsdir is not None:
        docsdir = [path.strip() for path in docsdir.strip().split('\n')]
    if workbook_filename:
        ignorable(docsdir, load_test_only)  # The commented-out reference to load_workbook below would use these vars.
        raise NotImplementedError("The load_workbook option is not implemented.")
        # TODO: load_workbook is undefined. where is it supposed to come from? -kmp 7-Aug-2022
        #       The definition would be in fourfront or cgap in src/encoded/__init__.py, which has a function
        #       similar to our 'main' here but with more functionality. Need to rethink that modularity so there
        #       is not code duplication and code that won't work. Could backport load_workbook to here, for example.
        #       -kmp 15-Sep-2022
        # load_workbook(app, workbook_filename, docsdir, test=load_test_only)

    # TODO: Maybe we should keep better track of what settings are used and not, and warn about unused options?
    return app



import os
import subprocess

from dcicutils.misc_utils import override_environ

from .postgresql_fixture import DEFAULT_SNOVAULT_DB_TEST_PORT


def test_snovault_db_test_port():

    # Will made this point during code review (18-Jun-2020):
    #  This makes virtualenv a requirement for successfully running the tests.
    #  Maybe that's something we want to do, but wanted to point it out
    #  just in case.
    command1 = "source %s/bin/activate" % os.environ['VIRTUAL_ENV']
    command2 = ("python -c 'from snovault.tests.postgresql_fixture import SNOVAULT_DB_TEST_PORT;"
                " print(SNOVAULT_DB_TEST_PORT)'")

    print("os.environ.get('SNOVAULT_DB_TEST_PORT') = %r" % os.environ.get('SNOVAULT_DB_TEST_PORT'))

    def _check_db_port():
        # We need to load the module freshly to find out if this setting is working, so we ask a subprocess.
        command = ["bash", "-c", '%s && %s' % (command1, command2)]
        output = subprocess.check_output(command)
        return output.decode('utf-8').strip()

    with override_environ(SNOVAULT_DB_TEST_PORT=None):
        actual_port = _check_db_port()
        default_port = str(DEFAULT_SNOVAULT_DB_TEST_PORT)
        print('actual_port=', actual_port)
        print('default_port=', default_port)
        assert actual_port == default_port

    custom_port = "314159"
    with override_environ(SNOVAULT_DB_TEST_PORT=custom_port):
        actual_port = _check_db_port()
        print('actual_port=', actual_port)
        print('custom_port=', custom_port)
        assert actual_port == custom_port



import re
import structlog
from copy import deepcopy
from collections import OrderedDict
from pyramid.httpexceptions import HTTPBadRequest
from urllib.parse import urlencode
from snovault import TYPES
from snovault.elasticsearch.create_mapping import determine_if_is_date_field
from .search_utils import (
    find_nested_path,  # convert_search_to_dictionary,
    QueryConstructionException,
    COMMON_EXCLUDED_URI_PARAMS, QUERY, FILTER, MUST, MUST_NOT, BOOL, MATCH, SHOULD,
    EXISTS, FIELD, NESTED, PATH, TERMS, RANGE, AGGS,  # REVERSE_NESTED,
    STATS,
    schema_for_field, get_query_field, search_log, MAX_FACET_COUNTS,
)


log = structlog.getLogger(__name__)


class LuceneBuilder:
    """ Collection of methods for working with Lucene queries. These operations can be used
        independently of the SearchBuilder state. See SearchBuilder for how these are used.

        Main points of entry:
            1. build_filters (construct the search query itself)
            2. build_facets (construct aggregations on search)
            3. verify_search_has_permissions (to be sure we did not strip permissions while building search)

        All other methods in this class are helper methods. Static methods are "leaf" operations that do
        not require additional function calls. Class methods call other methods within the class but could
        be "entry-point" methods as well.
    """
    to_from_pattern = re.compile("^(.*)[.](to|from)$")
    RANGE_DIRECTIONS = ['gt', 'gte', 'lt', 'lte']
    SMALLEST_NONZERO_IEEE_32 = 1.1754e-38  # smallest epsilon > 0 (estimate)
    SMALLEST_NEGATIVE_IEEE_32 = -3.4028e38
    # ref: http://www.cs.uwm.edu/classes/cs315/Bacon/Lecture/HTML/ch04s17.html
    # 1.00000000000000000000001 x 2^-127 = 1.1754e-38

    @staticmethod
    def apply_range_filters(range_filters, must_filters, es_mapping):
        """
        Applies the range filters to the 'must' subquery
        Tuple format is required to handle nested fields that are non-range (it is discarded in this case)
        Nested range fields must also be separated from other nested sub queries - see comment in handle_nested_filters
        Modifies must_filters in place

        :param range_filters: intermediary range_filter format to be converted to valid lucene
        :param must_filters: must_filters from build_sub_queries, this is where range filters are applied
        :param es_mapping: mapping of the item we searching on, as the range filter could be on a nested field
        """

        # tuple format is required to handle nested fields that are non-range (it is discarded in this case)
        # nested range fields must also be separated from other nested sub queries - see comment in 'handle_nested_filters'
        for range_field, range_def in range_filters.items():
            nested_path = find_nested_path(range_field, es_mapping)
            range_query = {RANGE: {range_field: range_def}}
            if 'add_no_value' in range_def:
                del range_def['add_no_value']
                range_query = {
                    BOOL: {
                        SHOULD: [
                            range_query,
                            {BOOL: {MUST_NOT: {EXISTS: {FIELD: range_field}}}}
                        ]
                    }
                }
            if nested_path:
                # look for existing nested sub query - must add to it if it exists
                found = False
                for query_part in must_filters:
                    nested_queries = query_part.get(BOOL, {}).get(MUST, {})
                    for query in nested_queries:
                        if NESTED in query and query[NESTED][PATH] == nested_path:
                            query[NESTED][QUERY][BOOL][MUST].append(range_query)
                            found = True
                            break  # if we found a valid path, add it here and continue
                # if we never found a path, this is the only nested query on that path, so just add it as is
                if not found:
                    new_nested_query = {
                        NESTED: {
                            PATH: nested_path,
                            QUERY: range_query
                        }
                    }
                    must_filters.append(new_nested_query)
            else:
                must_filters.append(range_query)

    @staticmethod
    def handle_should_query(field_name, options):
        """
        Builds a lucene 'should' subquery for every option for the given field

        :param field_name: full path to field
        :param options: list of options for that field
            ex: field_name='embedded.files.file_size.raw', options=[20, 30, 40]

        :return: dsl-subquery that is effectively an OR of all options on the field. See SHOULD.
        """
        should_query = {BOOL: {SHOULD: {TERMS: {field_name: options}}}}
        return should_query

    @classmethod
    def build_sub_queries(cls, field_filters, es_mapping):
        """
        Builds queries based on several things:
            - What the query field is
            - If that field is nested
            - If we would like to see items who do not have a value for this field. These items will have 'No value'
              for the relevant field.
            - If it is a positive (must) or negative (must_not) query. This is the level by which these are separated.

        :param field_filters: Intermediary format to be converted to valid lucene based on the es_mapping
        :param es_mapping: mapping of the item we are searching on
        :return: 4 tuple consisting of (must_filters, must_not_filters, must_filters_nested, must_not_filters_nested)
        """
        must_filters = []
        must_not_filters = []
        must_filters_nested = []
        must_not_filters_nested = []

        for query_field, filters in field_filters.items():
            # if we are nested, we must construct the query differently than normal
            nested_path = find_nested_path(query_field, es_mapping)
            if nested_path is not None:
                query_field = query_field.replace('.properties', '')
                strip_raw = query_field.replace('.raw', '')

                # if we are searching on the nested field itself, we must do something "special"
                if nested_path == strip_raw:
                    query_field = strip_raw

                    # if searching on 'No Value' on a nested field, the query has to be written
                    # slightly differently - note that you cannot combine a 'No value' search with
                    # anything else on this field path
                    if filters['add_no_value'] is True:
                        should_arr = [{EXISTS: {FIELD: query_field}}]
                        must_not_filters_nested.append((query_field, should_arr))
                        continue

                # Build must/must_not sub-queries
                # Example:
                # {'bool': {'must': {'bool':
                #   {'should': [{'match': {'embedded.hg19.hg19_hgvsg.raw': 'NC_000001.11:g.12185956del'}},
                #               {'match': {'embedded.hg19.hg19_hgvsg.raw': 'NC_000001.11:g.11901816A>T'}}
                #               ]}}}}
                # This is a "normal" query that we must convert to a "nested" sub-query on nested_path
                must_terms = cls.construct_nested_sub_queries(query_field, filters, key='must_terms')
                must_not_terms = cls.construct_nested_sub_queries(query_field, filters, key='must_not_terms')

                # XXX: In ES6, MUST -> MUST_NOT EXISTS does not work - have to use EXISTS under MUST_NOT
                # This means you cannot search on field=value or field DNE
                if filters['add_no_value'] is True:  # when searching on 'No Value'
                    should_arr = [must_not_terms] if must_not_terms else []
                    should_arr.append({BOOL: {MUST: {EXISTS: {FIELD: query_field}}}})  # field=value OR field DNE
                    must_not_filters_nested.append((query_field, should_arr))
                    if must_terms:
                        must_filters_nested.append((query_field, must_terms))
                else:  # when not searching on 'No Value'
                    should_arr = [must_terms] if must_terms else []
                    should_arr.append({EXISTS: {FIELD: query_field}})   # field=value OR field EXISTS
                    must_filters_nested.append((query_field, should_arr))
                    if must_not_terms:
                        must_not_filters_nested.append((query_field, must_not_terms))

            # if we are not nested, handle this with 'terms' query like usual
            else:
                must_terms = {TERMS: {query_field: filters['must_terms']}} if filters['must_terms'] else {}
                must_not_terms = {TERMS: {query_field: filters['must_not_terms']}} if filters['must_not_terms'] else {}
                if filters['add_no_value'] is True:
                    # add to must_not in an OR case, which is equivalent to filtering on 'No value'
                    should_arr = [must_terms] if must_terms else []
                    should_arr.append({BOOL: {MUST_NOT: {EXISTS: {FIELD: query_field}}}})  # field=value OR field DNE
                    must_filters.append((query_field, {BOOL: {SHOULD: should_arr}}))
                elif filters['add_no_value'] is False:
                    # add to must_not in an OR case, which is equivalent to filtering on '! No value'
                    should_arr = [must_terms] if must_terms else []
                    should_arr.append({EXISTS: {FIELD: query_field}})  # field=value OR field EXISTS
                    must_filters.append((query_field, {BOOL: {SHOULD: should_arr}}))
                elif must_terms:  # no filtering on 'No value'
                    must_filters.append((query_field, must_terms))
                if must_not_terms:
                    must_not_filters.append((query_field, must_not_terms))

        return must_filters, must_not_filters, must_filters_nested, must_not_filters_nested

    @staticmethod
    def construct_nested_sub_queries(query_field, filters, key='must_terms'):
        """
         Helper for build_sub_queries that constructs the base layer of sub-queries
         Note that due to the query structure, 'must' is always needed in the base level query,
         since at this point we have already split into 'must' or 'must_not'.

        :param query_field: field that we are querying
        :param filters: all filters
        :param key: one of 'must' or 'must_not'
        :return: a lucene sub-query filtering the query field based on the given filters
        :raises: QueryConstructionException if bad params make it here
        """
        if key not in ['must_terms', 'must_not_terms']:
            raise QueryConstructionException(
                query_type='nested',
                func='construct_nested_sub_queries',
                msg='Tried to handle nested filter with key other than must/must_not: %s' % key
            )
        my_filters = filters.get(key, [])
        if len(my_filters) == 0:
            return {}
        elif len(my_filters) == 1:  # see standard bool/match query
            return {MATCH: {query_field: my_filters[0]}}
        else:
            sub_queries = {BOOL: {SHOULD: []}}  # combine all options under SHOULD
            for option in my_filters:
                sub_queries[BOOL][SHOULD].append({MATCH: {query_field: option}})
            return sub_queries

    @classmethod
    def extract_field_from_to(cls, query_part):
        """ Neat helper method provided by Kent to clean up a step in 'handle_range_filters'.
            Extracts the field_name and whether it is a 'from' or 'to' query

        :param query_part: query part to parse, such as "field.a.from" or "field.to". See the regexp.
        :return: 3-tuple consisting of whether or not there was a match, the first grouping and the second grouping
                 ie: (True, field.name, 'from')
        """
        match = cls.to_from_pattern.match(query_part)
        if match is not None:
            return bool(match), match.group(1), match.group(2)
        return False, None, None

    @classmethod
    def canonicalize_bounds(cls, range_filter):
        """ Canonicalizes the bounds of the range filter such that they are
            inclusive on the lower bound and exclusive on the upper bound.
        """
        lower, upper = -1e38, 1e38  # very large numbers that should never be in range
        for direction, pivot in range_filter.items():
            pivot = float(pivot)
            if direction == 'lte':
                upper = pivot + cls.SMALLEST_NONZERO_IEEE_32
            elif direction == 'lt':
                upper = pivot
            elif direction == 'gte':
                lower = pivot
            elif direction == 'gt':
                lower = pivot - cls.SMALLEST_NONZERO_IEEE_32
        return lower, upper

    @classmethod
    def range_includes_zero(cls, range_filter):
        """ Returns True if the given range_filter includes the value 0. """
        lower, upper = cls.canonicalize_bounds(range_filter)
        return lower <= 0 <= upper

    @classmethod
    def handle_range_filters(cls, request, result, field_filters, doc_types):
        """
        Constructs range_filters based on the given filters as part of the MUST sub-query

        :param request: current request
        :param result: result to modify in place
        :param field_filters: filters to look at
        :param doc_types: types we are searching on
        :return: constructed range_filters
        """
        range_filters = {}

        for field, term in request.normalized_params.items():
            not_field = False  # keep track if query is NOT (!)
            exists_field = False  # keep track of null values
            range_type = False  # If we determine is a range request (field.to, field.from), will be populated with string 'date' or 'numerical'
            range_direction = None
            field_schema = {}
            if field == 'q' or field in COMMON_EXCLUDED_URI_PARAMS:
                continue
            elif field == 'type' and term != 'Item':
                continue
            elif term == 'No value':
                exists_field = True

            # Check for date or numerical range filters
            is_range, f_field, which = cls.extract_field_from_to(field)
            if is_range:
                if which == 'to':
                    range_direction = 'lte'
                else:
                    range_direction = 'gte'

                # If schema for field is not found (and range_type thus not set),
                # then treated as ordinary term filter (likely will get 0 results)
                field_schema = schema_for_field(f_field, request, doc_types)
                if field_schema:
                    range_type = 'date' if determine_if_is_date_field(f_field, field_schema) else 'numerical'

            # Add filter to result
            qs = urlencode([
                (k.encode('utf-8'), v.encode('utf-8'))
                for k, v in request.normalized_params.items()
                if k != "limit" and k != "from" and not (k == field and v == term)
            ])
            remove_path = '{}?{}'.format(request.path, qs)

            # default to searching type=Item rather than empty filter path
            if remove_path[-1] == '?':
                remove_path += 'type=Item'

            result['filters'].append({
                'field': field,
                'term': term,
                'remove': remove_path
            })

            # handle NOT
            if field.endswith('!'):
                field = field[:-1]
                not_field = True

            # Add filter to query
            if range_type and f_field and range_type in ('date', 'numerical'):
                query_field = 'embedded.' + f_field
            elif field.startswith('validation_errors') or field.startswith('aggregated_items'):
                query_field = field + '.raw'
            elif field == 'type':
                query_field = 'embedded.@type.raw'
            else:
                query_field = 'embedded.' + field + '.raw'

            if range_type:
                if query_field not in range_filters:
                    range_filters[query_field] = {}
                    if range_type == 'date':
                        range_filters[query_field]['format'] = 'yyyy-MM-dd HH:mm'

                if range_direction in cls.RANGE_DIRECTIONS:
                    if range_type == "date" and len(term) == 10:  # TODO: refactor to use regex -Will 06/24/2020
                        # Correct term to have hours, e.g. 00:00 or 23:59, if not otherwise supplied.
                        if range_direction == 'gt' or range_direction == 'lte':
                            term += ' 23:59'
                        elif range_direction == 'gte' or range_direction == 'lt':
                            term += ' 00:00'

                    if range_filters[query_field].get(range_direction) is None:
                        range_filters[query_field][range_direction] = term
                    else:
                        # If have a value already (e.g. multiple ranges selected), choose the widening option.
                        if range_direction == 'gt' or range_direction == 'gte':
                            if term < range_filters[query_field][range_direction]:
                                range_filters[query_field][range_direction] = term
                        elif range_direction == 'lt' or range_direction == 'lte':
                            if term > range_filters[query_field][range_direction]:
                                range_filters[query_field][range_direction] = term

                # Check if schema requests no value
                if 'items' in field_schema:  # we are searching on an array of numerics
                    field_schema = field_schema['items']
                if field_schema.get('add_no_value', False) and cls.range_includes_zero(range_filters[query_field]):
                    range_filters[query_field]['add_no_value'] = True

            # add these to field_filters directly, handle later with build_sub_queries
            else:
                if query_field not in field_filters:
                    field_filters[query_field] = {
                        'must_terms': [],
                        'must_not_terms': [],
                        'add_no_value': None
                    }

                # handle case of filtering for null values
                if exists_field:
                    # the value below is True when we want to include 'No value' as a filter
                    field_filters[query_field]['add_no_value'] = False if not_field else True
                    continue

                if not_field:
                    field_filters[query_field]['must_not_terms'].append(term)
                else:
                    field_filters[query_field]['must_terms'].append(term)

        return range_filters

    @staticmethod
    def initialize_field_filters(request, principals, doc_types):
        """ Helper function for build_filters
            Initializes field filters with filters that exist on all searches, does some basic updates
        """
        field_filters = {
            'principals_allowed.view': {
                'must_terms': principals,
                'must_not_terms': [],
                'add_no_value': None
            },
            'embedded.@type.raw': {
                'must_terms': doc_types,
                'must_not_terms': [],
                'add_no_value': None
            },
            'embedded.status.raw': {
                'must_terms': [],
                'must_not_terms': [],
                'add_no_value': None
            }
        }

        # Exclude status=deleted Items unless explicitly requested/filtered-in.
        if 'deleted' not in request.normalized_params.getall('status'):
            field_filters['embedded.status.raw']['must_not_terms'].append('deleted')
        if 'replaced' not in request.normalized_params.getall('status'):
            field_filters['embedded.status.raw']['must_not_terms'].append('replaced')

        # Exclude type=TrackingItem and type=OntologyTerm from results unless are explictly specified
        if 'TrackingItem' not in doc_types:
            field_filters['embedded.@type.raw']['must_not_terms'].append('TrackingItem')
        if 'OntologyTerm' not in doc_types:
            field_filters['embedded.@type.raw']['must_not_terms'].append('OntologyTerm')

        return field_filters

    @staticmethod
    def build_nested_query(nested_path, query):
        """ Takes the given query and converts it into a nested query on the
            given path.
        """
        return {
            NESTED: {
                PATH: nested_path,
                QUERY: query
            }
        }

    @classmethod
    def handle_nested_filters_v2(cls, must_nested_filters, must_not_nested_filters, es_mapping):
        """ This function implements nested query construction.

            When building a nested query, unlike with traditional queries, selections on the same
            field must occur in the same nested sub-query in order to be applied as an intersect
            condition on the object field. Previously we would create separate nested sub-queries
            per field selection, which causes them to be OR'd.

            :param must_nested_filters: conditions filtered on in the affirmative
            :param must_not_nested_filters: conditions filtered on in the negative
            :param es_mapping: the ES mapping of the type we are searching on
            :returns: a nested sub-query that can be added directly to the parent query
        """
        # Build base query structure
        # always use MUST + sub queries for MUST_NOT
        nested_query = {
            BOOL: {
                MUST: [],
                MUST_NOT: [],
            }
        }

        # Maps a nested path to a 2-tuple of it's key (must/must_not) and index
        nested_path_to_index_map = {}

        # Build array of key, (field, query) so we can process the filters in a single pass
        # note that MUST queries are always processed first
        filters_to_work_on = []
        if must_nested_filters:
            filters_to_work_on += zip([MUST] * len(must_nested_filters), must_nested_filters)
        if must_not_nested_filters:
            filters_to_work_on += zip([MUST_NOT] * len(must_not_nested_filters), must_not_nested_filters)

        # Process key (must/must_not), field (target of search), query (condition)
        # iteratively building nested_query
        for key, (field, query) in filters_to_work_on:
            nested_path = find_nested_path(field, es_mapping)

            # if we've never seen this path before, bootstrap a sub-query for it
            if nested_path not in nested_path_to_index_map:

                # set in tracking, note it is order dependent
                new_index = len(nested_query[BOOL][key])
                nested_path_to_index_map[nested_path] = (key, new_index)

                # this nested path could have more filters (under differing keys)
                # bootstrap an entire sub-query for this path
                combined_query = {
                    BOOL: {
                        MUST: [],
                        MUST_NOT: []
                    }
                }
                for sub_query in query:
                    # Special case for EXISTS, since we cannot construct these like normal
                    # queries - add DOES NOT EXIST queries to MUST branches, as these are
                    # automatically added to MUST_NOT branch
                    if EXISTS in sub_query and key == MUST_NOT:
                        combined_query[BOOL][MUST].append(sub_query)
                    elif sub_query.get(BOOL, {}).get(MUST, {}).get(EXISTS, None):
                        combined_query[BOOL][MUST].append(sub_query)
                    else:
                        combined_query[BOOL][key].append(sub_query)

                # add the combined_query for this nested path to the global nested query
                nested_query[BOOL][key].append(cls.build_nested_query(nested_path, combined_query))

            # We have seen this nested_path before, so in order to achieve proper intersect
            # behavior all conditions must be present on the same nested sub-query
            else:

                # extract the location of the nested query we would like to add to
                # note that the key under which the previous query was added could differ
                # from the key we are seeing now ie: EXIST (must) combined with != (must_not)
                prev_key, path_index = nested_path_to_index_map[nested_path]
                leaf_query = nested_query[BOOL][prev_key][path_index][NESTED][QUERY][BOOL][key]

                # leaf_query is the sub-query we want to build off of
                # its possible our current query contains multiple conditions
                if isinstance(query, list):
                    leaf_query += query
                elif isinstance(query, dict):
                    leaf_query.append(query)
                else:
                    raise QueryConstructionException(
                        query_type='nested',
                        func='handle_nested_filters_v2',
                        msg='passed a query with a bad type: %s' % query
                    )

        return nested_query

    @classmethod
    def build_filters(cls, request, query, result, principals, doc_types, es_mapping):
        """
        This function builds the Elasticsearch query based on the request. The structure of the query
        is approximately represented below. 'Approximate' because you could not copy-paste directly into
        Lucene, but should illustrate enough so it is comprehensible. Note the 'nested' nature of the query.

        QUERY HIERARCHY ('approximate' lucene syntax):
            {
                'query': {
                    'bool': {
                        'filter': {
                            'bool': {
                                'must': {
                                    <positive+range sub-queries>
                                    +
                                    'bool': {
                                        'should': { option1, option2 ... }
                                    }
                                },
                                'must_not': {
                                    <negative sub-queries>
                                    +
                                    'bool': {
                                        'should': { option1, option2 ...
                                }}}}}
            }
        There are several different sub-queries, but most of them are built using 'bool' -> 'must'.
        A brief overview follows.
            * nested data type sub-queries have a special format. See 'handle_nested_filters'.
            * range filter sub-queries also have their own special format. See 'apply_range_filters'. Note that
                the format is extra special when you're applying a range filter to a nested data type.
            * 'terms' filters are what we 'normally' use.

        :param request: Current request
        :param query: Current search query body
        :param result: Response to be returned from the view ('/search')
        :param principals: Active user roles
        :param doc_types: Document type we are searching on
        :param es_mapping: Elasticsearch mapping of the document type we're searching on
        :returns: 2-tuple containing the updated search based on the request parameters and
                  information on the filters used in the query.
        """

        # these next two dictionaries should each have keys equal to query_field
        # and values: must_terms: [<list of terms>], must_not_terms: [<list of terms>], add_no_value: True/False/None
        field_filters = cls.initialize_field_filters(request, principals, doc_types)
        range_filters = cls.handle_range_filters(request, result, field_filters, doc_types)

        # construct queries
        must_filters, must_not_filters, \
        must_filters_nested, must_not_filters_nested = cls.build_sub_queries(field_filters, es_mapping)

        # initialize filter hierarchy
        final_filters = {BOOL: {MUST: [f for _, f in must_filters], MUST_NOT: [f for _, f in must_not_filters]}}

        # Build nested queries
        final_nested_query = cls.handle_nested_filters_v2(must_filters_nested, must_not_filters_nested, es_mapping)
        final_filters[BOOL][MUST].append(final_nested_query)

        # add range filters after (so nested ranges can be applied with existing nested queries)
        cls.apply_range_filters(range_filters, final_filters[BOOL][MUST], es_mapping)

        # at this point, final_filters is valid lucene and can be dropped into the query directly
        query[QUERY][BOOL][FILTER] = final_filters
        return query, final_filters

    @staticmethod
    def _check_and_remove(compare_field, facet_filters, active_filter, query_field, filter_type):
        """ Does the actual 'check and removal' since this code is duplicated throughout. """
        if compare_field == query_field:
            facet_filters[filter_type].remove(active_filter)
            return True
        return False

    @classmethod
    def _check_and_remove_terms(cls, facet_filters, active_filter, query_field, filter_type):
        """ Helper function for _remove_from_active_filters that handles filter removal for terms query """
        # there should only be one key here
        for compare_field in active_filter[TERMS].keys():
            # remove filter for a given field for that facet
            # skip this for type facet (field = 'type')
            # since we always want to include that filter.
            if (query_field != 'embedded.@type.raw' and  # this evaluation order MUST be preserved!
                    cls._check_and_remove(compare_field, facet_filters, active_filter, query_field, filter_type)):
                break

    @classmethod
    def _check_and_remove_range(cls, facet_filters, active_filter, query_field, filter_type):
        """ Helper function for _remove_from_active_filters that handles filter removal for terms query """
        for compare_field in active_filter[RANGE].keys():
            if cls._check_and_remove(compare_field, facet_filters, active_filter, query_field, filter_type):
                break

    @classmethod
    def _check_and_remove_bool_should(cls, facet_filters, active_filter, query_field, filter_type):
        """ Helper function for _remove_from_active_filters that handles filter removal for boolean queries that
            have multiple options (inner SHOULD query)
        """
        # handle No value case
        inner_bool = None
        inner_should = active_filter.get(BOOL).get(SHOULD, [])
        for or_term in inner_should:
            # this may be naive, but assume first non-terms
            # filter is the No value query
            if TERMS in or_term or RANGE in or_term:
                continue
            else:
                inner_bool = or_term
                break
        if EXISTS in inner_bool:
            compare_field = inner_bool[EXISTS].get(FIELD)
        else:
            # attempt to get the field from the alternative No value syntax
            compare_field = inner_bool.get(BOOL, {}).get(MUST_NOT, {}).get(EXISTS, {}).get(FIELD)
        if query_field != 'embedded.@type.raw':
            cls._check_and_remove(compare_field, facet_filters, active_filter, query_field, filter_type)

    @classmethod
    def _check_and_remove_match_from_should(cls, query_options, facet_filters, active_filter, query_field,
                                            filter_type):
        """ Helper function that searches a MATCH query for the given query_field, removing the
            active filter if found.
        """
        for inner_query in query_options:
            if MATCH in inner_query:
                for field in inner_query.get(MATCH, {}).keys():  # should be only one per block
                    if cls._check_and_remove(field, facet_filters, active_filter, query_field,
                                             filter_type):
                        return
            else:
                search_log(log_handler=log, msg='Encountered a unexpected nested structure in '
                                                'query: %s' % inner_query)

    @classmethod
    def _check_and_remove_nested(cls, facet_filters, active_filter, query_field, filter_type):
        """ Helper function for _remove_from_active_filters that handles filter removal for nested query
            Reminder that this code is responsible for constructing the aggregation filter, hence the desire
            to omit selections on the field we are aggregating on.
        """
        nested_sub_query = active_filter[NESTED][QUERY]

        # For No value searches
        if EXISTS in nested_sub_query:
            field = nested_sub_query.get(EXISTS, {}).get(FIELD)
            cls._check_and_remove(field, facet_filters, active_filter, query_field, filter_type)

        # For all other searches
        elif BOOL in nested_sub_query:
            for inner_filter_type in [MUST, MUST_NOT]:
                for nested_option in nested_sub_query[BOOL].get(inner_filter_type, []):
                    if isinstance(nested_option, dict):

                        # For structure like this:
                        #   {'bool': {'must': [{'match': {'embedded.hg19.hg19_hgvsg.raw': 'NC_000001.11:g.12185956del'}]
                        if MATCH in nested_option:
                            for field in nested_option.get(MATCH, {}).keys():  # should only be one per block
                                if cls._check_and_remove(field, facet_filters, active_filter, query_field, filter_type):
                                    break

                        # For structure like this:
                        # {'bool': {'should':
                        # [{'match': {'embedded.variant.genes.genes_most_severe_consequence.coding_effect.raw':
                        #       'Missense'}},
                        # {'match': {'embedded.variant.genes.genes_most_severe_consequence.coding_effect.raw':
                        #       'Synonymous'}}]}}
                        elif BOOL in nested_option:
                            inner_inner_bool = nested_option[BOOL]
                            if SHOULD in inner_inner_bool:
                                cls._check_and_remove_match_from_should(inner_inner_bool[SHOULD], facet_filters,
                                                                        active_filter,
                                                                        query_field, filter_type)

                    # For structure like this:
                    #   {'bool': {'must': {'bool': {'should':
                    #       [{'match': {'embedded.hg19.hg19_hgvsg.raw': 'NC_000001.11:g.12185956del'}},
                    #       {'match': {'embedded.hg19.hg19_hgvsg.raw': 'NC_000001.11:g.11901816A>T'}}]}}}}
                    elif isinstance(nested_option, str):
                        inner_bool = nested_sub_query[BOOL].get(inner_filter_type, {})
                        if SHOULD in inner_bool:
                            cls._check_and_remove_match_from_should(inner_bool[SHOULD], facet_filters, active_filter,
                                                                    query_field, filter_type)

                        # For structure like this:
                        # {'bool': {'should': [
                        #    {'match': {'embedded.variant.genes.genes_most_severe_consequence.impact.raw': 'MODIFIER'}},
                        #    {'match': {'embedded.variant.genes.genes_most_severe_consequence.impact.raw': 'LOW'}}]}}
                        elif BOOL in inner_bool:
                            inner_inner_bool = inner_bool[BOOL]
                            if SHOULD in inner_inner_bool:
                                cls._check_and_remove_match_from_should(inner_inner_bool[SHOULD], facet_filters, active_filter,
                                                                        query_field, filter_type)

                    else:
                        search_log(log_handler=log, msg='Encountered a unexpected nested structure at top level: %s'
                                                        % nested_sub_query[BOOL])

    @classmethod
    def _remove_from_active_filters(cls, facet_filters, query_field, active_filter, filter_type):
        """ Helper function for generate_filters_for_terms_agg_from_search_filters
            Modifies facet_filters in place to remove the active_filter if it matches
            the given query field.
            This function is intended to be called on every "sub part" of the base query
            for every aggregation.

            TODO: Optimize this - it is inefficient application side regardless of the dominating ES cost

            :param facet_filters: intended filter block to be used on the aggregation
            :param query_field: field that we are aggregating on
            :param active_filter: which "sub part" of the facet filters we are examining
            :param filter_type: one of MUST or MUST_NOT
        """
        if BOOL in active_filter and SHOULD in active_filter[BOOL]:
            cls._check_and_remove_bool_should(facet_filters, active_filter, query_field, filter_type)
        elif TERMS in active_filter:
            cls._check_and_remove_terms(facet_filters, active_filter, query_field, filter_type)
        elif RANGE in active_filter:
            cls._check_and_remove_range(facet_filters, active_filter, query_field, filter_type)
        elif NESTED in active_filter:
            cls._check_and_remove_nested(facet_filters, active_filter, query_field, filter_type)

    @classmethod
    def generate_filters_for_terms_agg_from_search_filters(cls, query_field, search_filters, string_query):
        """
        We add a copy of our filters to each facet, minus that of
        facet's field itself so that we can get term counts for other terms filters.
        And be able to filter w/ it.
        Remove filters from fields they apply to.
        For example, the 'biosource_type' aggs should not have any
        biosource_type filter in place.
        Handle 'must' and 'must_not' filters separately

        :param query_field: field terms agg is on
        :param search_filters: intermediary format prior to any valid lucene representing the search_filters
                               from the front-end
        :param string_query: query string if provided
        :return: Copy of search_filters, minus filter for current query_field (if one set).
        """
        if not search_filters or BOOL not in search_filters:  # a sane default if this happens -Will 11/17/20
            log.error('Encountered an unexpected query format: %s' % search_filters)
            return {BOOL: {MUST: [{TERMS: {'principals_allowed.view': ['system.Everyone']}}]}}

        facet_filters = deepcopy(search_filters[BOOL])

        for filter_type in [MUST, MUST_NOT]:
            # active_filter => e.g. { 'terms' : { 'embedded.@type.raw': ['ExperimentSetReplicate'] } }
            for active_filter in search_filters[BOOL][filter_type]:
                cls._remove_from_active_filters(facet_filters, query_field, active_filter, filter_type)

        # add the string_query, if present, to the bool term with facet_filters
        if string_query and string_query[MUST]:
            # combine statements within 'must' for each
            facet_filters[MUST].append(string_query[MUST])

        return {BOOL: facet_filters}

    @staticmethod
    def set_additional_aggregations(search_as_dict, request, doc_types, extra_aggregations=None):
        """
        Per-type aggregations may be defined in schemas. Apply them OUTSIDE of globals so they act on our
        current search filters. Warning: `search_as_dict` is modified IN PLACE.

        :param search_as_dict: elasticsearch_dsl object converted to_dict()
        :param request: current request
        :param doc_types: types we are searching on
        :param extra_aggregations: aggregations to add
        :return: search_as_dict, same as originally passed in, but modified in this function
        """

        types = request.registry[TYPES]
        schema = types[doc_types[0]].schema

        if schema.get('aggregations'):
            for schema_agg_name in schema['aggregations'].keys():
                if schema_agg_name == 'all_items':
                    raise QueryConstructionException(
                        query_type='aggregations',
                        func='set_additional_aggregations',
                        msg='all_items is a reserved agg name and not allowed as an extra aggregation name.')
                search_as_dict['aggs'][schema_agg_name] = schema['aggregations'][schema_agg_name]

        if extra_aggregations:
            for extra_agg_name in extra_aggregations.keys():
                if extra_agg_name == 'all_items':
                    raise QueryConstructionException(
                        query_type='extra_aggregations',
                        func='set_additional_aggregations',
                        msg='all_items is a reserved agg name and not allowed as an extra aggregation name.')
                search_as_dict['aggs'][extra_agg_name] = extra_aggregations[extra_agg_name]

        return search_as_dict

    @staticmethod
    def _build_nested_aggregation(sub_query, nested_path, requested=None):
        """ Builds a nested aggregation.

            :param sub_query: query to use as the 'primary_agg'
            :param nested_path: path to nested object we are searching on
            :param requested: requested agg, if any
            :returns: the nested form of sub_query
        """
        if requested:
            return {
                NESTED: {
                    PATH: nested_path
                },
                AGGS: {
                    'primary_agg': sub_query,
                    'requested_agg': requested
                }
            }
        else:
            return {
                NESTED: {
                    PATH: nested_path
                },
                AGGS: {
                    'primary_agg': sub_query
                }
            }

    @classmethod
    def _add_stats_aggregation(cls, field, facet, field_schema, query_field, search_filters, string_query,
                               nested_path, aggs, agg_name):
        """ Builds a stats aggregation, adding it to the given aggs.

            :param field: raw field name we are searching on (ex: AF)
            :param facet: facet metadata
            :param field_schema: schema for the field we are searching on
            :param query_field: ES document field we are searching on (ex: embedded.AF)
            :param search_filters: filters we are searching on
            :param string_query: simple query string if specified
            :param nested_path: path to nested object we are aggregating on
            :param aggs: the aggregation object we are building
            :param agg_name: name of the aggregation we are building
        """
        is_date_field = field_schema and determine_if_is_date_field(field, field_schema)
        is_numerical_field = field_schema and field_schema['type'] in ("integer", "float", "number")
        if is_date_field:
            facet['field_type'] = 'date'
        elif is_numerical_field:
            facet["field_type"] = field_schema['type'] or "number"
            if "number_step" not in facet:
                if "number_step" in field_schema:
                    facet["number_step"] = field_schema['number_step']
                elif facet["field_type"] == "integer":
                    facet["number_step"] = 1
                else:  # Default
                    facet["number_step"] = "any"
        facet_filters = cls.generate_filters_for_terms_agg_from_search_filters(query_field, search_filters,
                                                                               string_query)
        # stats aggregations could be nested too
        stats_agg = {
            STATS: {
                'field': query_field
            }
        }
        if nested_path:
            facet['aggregation_type'] = 'nested:stats'
            aggs[facet['aggregation_type'] + ':' + agg_name] = {
                AGGS: {
                    'primary_agg': cls._build_nested_aggregation(stats_agg, nested_path)
                },
                FILTER: facet_filters
            }

        else:
            aggs[facet['aggregation_type'] + ":" + agg_name] = {
                AGGS: {
                    'primary_agg': stats_agg
                },
                FILTER: facet_filters
            }

    @classmethod
    def _build_range_aggregation(cls, query_field, ranges):
        """ Builds a range aggregation.
            Detects when 0-0 range is specified and replaces 'to' with the
            smallest IEEE 32 value such that the bucket effectively only captures
            the value 0.
        """
        for r in ranges:
            if 'from' in r and 'to' in r:
                if r['from'] == 0 and r['to'] == 0:
                    r['to'] = cls.SMALLEST_NONZERO_IEEE_32
            if 'to' in r and r['to'] != cls.SMALLEST_NONZERO_IEEE_32:
                r['to'] += cls.SMALLEST_NONZERO_IEEE_32
        return {
            RANGE: {
                FIELD: query_field,
                'ranges': ranges
            }
        }

    @classmethod
    def _add_range_aggregation(cls, facet, query_field, search_filters, string_query, nested_path, aggs, agg_name):
        """ Builds a range aggregation utilizing the ranges defined on schema facet, adding it to the given aggs.

            :param facet: facet metadata
            :param query_field: field we are searching on.
            :param search_filters: filters we are searching on
            :param string_query: simple query string if specified
            :param nested_path: path to nested object we are aggregating on
            :param aggs: the aggregation object we are building
            :param agg_name: name of the aggregation we are building
        """
        facet_filters = cls.generate_filters_for_terms_agg_from_search_filters(query_field, search_filters,
                                                                               string_query)
        ranges = [{k: v for k, v in r.items() if k in ['from', 'to']} for r in facet['ranges']]
        range_agg = cls._build_range_aggregation(query_field, ranges)
        if nested_path:
            facet['aggregation_type'] = 'nested:range'
            field = facet['aggregation_type'] + ':' + agg_name
            range_agg = cls._build_nested_aggregation(range_agg, nested_path)
        else:
            facet['aggregation_type'] = RANGE
            field = facet['aggregation_type'] + ':' + agg_name
        aggs[field] = {
            AGGS: {
                'primary_agg': range_agg
            },
            FILTER: facet_filters
        }

    @staticmethod
    def _build_terms_aggregation(query_field, facet, requested_values=None, nested=False):
        """ Builds a terms aggregation, specifically requesting counts for any selected values. """
        agg = {
            TERMS: {
                'size': MAX_FACET_COUNTS,
                'field': query_field,
                'missing': facet.get('missing_value_replacement', 'No value')
            }
        }
        if requested_values:  # getall returns [], not None
            agg[TERMS]['include'] = requested_values
        if nested:
            agg[AGGS] = {
                'primary_agg_reverse_nested': {
                    'reverse_nested': {}
                }
            }
        return agg

    @classmethod
    def _add_terms_aggregation(cls, facet, query_field, search_filters, string_query, nested_path, aggs, agg_name,
                               requested_values):
        """ Builds a standard terms aggregation, setting a nested identifier to be repaired later
            by elasticsearch_dsl, adding it to the given aggs.

            :param facet: facet metadata
            :param query_field: field we are searching on.
            :param search_filters: filters we are searching on
            :param string_query: simple query string if specified
            :param nested_path: path to nested object we are aggregating on
            :param aggs: the aggregation object we are building
            :param agg_name: name of the aggregation we are building
            :param requested_values: values for this terms agg we requested (to be explicitly included)
        """
        is_nested = nested_path is not None
        if is_nested:
            facet['aggregation_type'] = NESTED  # special in that it is used to identify (broken) facets - Will 11/17/20
        else:
            facet['aggregation_type'] = TERMS

        facet_filters = cls.generate_filters_for_terms_agg_from_search_filters(query_field, search_filters,
                                                                               string_query)
        terms_aggregation = cls._build_terms_aggregation(query_field, facet, None, is_nested)

        # NOTE: if we requested values for this field, we must expand to do two aggregations
        # Unfortunately when you pass "include" to a terms aggregation it acts as a hard filter,
        # not a "force bucket", which makes implementing this very tricky. To get around this we
        # expand to 2 aggregations - one for the requested field and one for the remaining top fields
        if requested_values:
            terms_aggregation_requested = cls._build_terms_aggregation(query_field, facet, requested_values,
                                                                       is_nested)
            if nested_path:
                aggs[facet['aggregation_type'] + ":" + agg_name] = {
                    AGGS: {'primary_agg':
                               cls._build_nested_aggregation(terms_aggregation, nested_path,
                                                             terms_aggregation_requested),
                    },
                    FILTER: facet_filters,
                }
            else:
                aggs[facet['aggregation_type'] + ":" + agg_name] = {
                    AGGS: {
                        'primary_agg': terms_aggregation_requested,
                        'requested_agg': terms_aggregation
                    },
                    FILTER: facet_filters,
                }

        else:
            if nested_path:
                aggs[facet['aggregation_type'] + ":" + agg_name] = {
                    AGGS: {'primary_agg':
                               cls._build_nested_aggregation(terms_aggregation, nested_path),
                    },
                    FILTER: facet_filters,
                }
            else:
                aggs[facet['aggregation_type'] + ":" + agg_name] = {
                    AGGS: {
                        'primary_agg': terms_aggregation
                    },
                    FILTER: facet_filters,
                }

    @classmethod
    def build_facets(cls, query, facets, search_filters, string_query, request, doc_types,
                     custom_aggregations=None, size=25, from_=0, es_mapping=None):
        """
        Sets facets in the query as ElasticSearch aggregations, with each aggregation to be
        filtered by search_filters minus filter affecting facet field in order to get counts
        for other facet term options.
        ES5 - simply sets aggs by calling update_from_dict after adding them in
            :param facets:         Facet field (0) in object dot notation, and a dict or OrderedDict with title property (1).
            :type  facets:         List of tuples.
            :param search_filters: Dict of filters which are set for the ES query in build_filters
            :param string_query:   Dict holding the query_string used in the search
        """
        if from_ != 0:
            return query

        aggs = OrderedDict()
        for field, facet in facets:  # E.g. 'type','experimentset_type','experiments_in_set.award.project', ...
            field_schema = schema_for_field(field, request, doc_types, should_log=True)
            query_field = get_query_field(field, facet)
            nested_path = find_nested_path(query_field, es_mapping)
            requested_values = request.params.getall(field)

            # Build the aggregation based on its type (by side-effect) - stats, range or terms
            agg_name = field.replace('.', '-')
            facet_type = facet.get('aggregation_type')
            if facet_type in ['stats', 'nested:stats']:
                cls._add_stats_aggregation(field, facet, field_schema, query_field, search_filters, string_query,
                                           nested_path, aggs, agg_name)
            elif facet_type in ['range', 'nested:range']:
                cls._add_range_aggregation(facet, query_field, search_filters, string_query, nested_path,
                                           aggs, agg_name)
            else:  # assume terms
                cls._add_terms_aggregation(facet, query_field, search_filters, string_query, nested_path,
                                           aggs, agg_name, requested_values)

            # Update facet with title, description from field_schema, if missing.
            if facet.get('title') is None and field_schema and 'title' in field_schema:
                facet['title'] = field_schema['title']
            if facet.get('description') is None and field_schema and 'description' in field_schema:
                facet['description'] = field_schema['description']

        # to achieve OR behavior within facets, search among GLOBAL results,
        # not just returned ones. to do this, wrap aggs in ['all_items']
        # and add "global": {} to top level aggs query
        # see elasticsearch global aggs for documentation (should be ES5 compliant)
        query['aggs'] = {
            'all_items': {
                'global': {},
                'aggs': aggs
            }
        }

        if size == 0:
            # Only perform aggs if size==0 requested, to improve performance for search page queries.
            # We do currently have (hidden) monthly date histogram facets which may yet to be utilized for common size!=0 agg use cases.
            cls.set_additional_aggregations(query, request, doc_types, custom_aggregations)

        return query

    @staticmethod
    def verify_search_has_permissions(request, query):
        """
        Inspects the search object to ensure permissions are still present on the query
        This method depends on the query structure defined in 'build_filters'.

        :param request: the current request
        :param query: search query object to inspect
        :raises: HTTPBadRequest if permissions not present
        """
        effective_principals_on_query = None
        found = False  # set to True if we found valid 'principals_allowed.view'
        try:
            for boolean_clause in [query['query']['bool']['filter']]:  # should always be present
                if 'bool' in boolean_clause and 'must' in boolean_clause['bool']:  # principals_allowed.view is on 'must'
                    possible_permission_block = boolean_clause['bool']['must']
                    for entry in possible_permission_block:
                        if 'terms' in entry:
                            if 'principals_allowed.view' in entry['terms']:
                                effective_principals_on_query = entry['terms']['principals_allowed.view']
                                if effective_principals_on_query != request.effective_principals:
                                    raise QueryConstructionException(
                                        query_type='principals',
                                        func='verify_search_has_permissions',
                                        msg='principals_allowed was modified - see application logs')
                                else:
                                    found = True
                                    break
        except QueryConstructionException:
            search_log(log_handler=log, msg='Detected URL query param manipulation, principals_allowed.view was'
                                            ' modified from %s to %s' % (request.effective_principals,
                                                                         effective_principals_on_query))
            raise HTTPBadRequest('The search failed - the DCIC team has been notified.')
        except KeyError:
            search_log(log_handler=log, msg='Malformed query detected while checking for principals_allowed')
            raise HTTPBadRequest('The search failed - the DCIC team has been notified.')
        if not found:
            search_log(log_handler=log, msg='Did not locate principals_allowed.view on search query body: %s'
                                            % query)
            raise HTTPBadRequest('The search failed - the DCIC team has been notified.')

    @classmethod
    def compound_search(cls, sub_queries, intersect=False):
        """  Takes an array of sub-queries and merges them into one query

        :param sub_queries: list of query to be combined, typically starting with "bool"
        :param intersect: whether or not to intersect the sub-queries
        :return: lucene query combining the sub_queries with OR
        """
        if not intersect:
            key = SHOULD
        else:
            key = MUST

        query = {
            'query': {
                'bool': {
                    key: []
                }
            }
        }
        for q in sub_queries:
            query['query']['bool'][key].append(q)
        return query



import argparse
import os

import logging
from dcicutils.deployment_utils import create_file_from_template
from dcicutils.misc_utils import ignored, PRINT, classproperty

EPILOG = __doc__


class PrepOptions:

    @classproperty
    def ROOT_DIR(cls):  # noQA - cls is the right var
        # return os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
        return os.path.abspath(os.curdir)

    @classproperty
    def DOCKER_COMPOSE_FILE(cls):  # noQA - cls is the right var
        return os.path.join(cls.ROOT_DIR, 'docker-compose.yml')

    @classproperty
    def DOCKER_DEVELOPMENT_INI_FILE(cls):  # noQA - cls is the right var
        return os.path.join(cls.ROOT_DIR, "deploy/docker/local/docker_development.ini")

    @classproperty
    def DEVELOPMENT_INI_FILE(cls):  # noQA - cls is the right var
        return os.path.join(cls.ROOT_DIR, "development.ini")

    @classproperty
    def TEST_INI_FILE(cls):  # noQA - cls is the right var
        return os.path.join(cls.ROOT_DIR, "test.ini")


DATA_SET_CHOICES = ['prod', 'test', 'local', 'deploy']
DEFAULT_DATA_SET = 'local'


def empty_assignment(line, expanded):
    ignored(line)
    return expanded.strip().endswith(': ""')


def template_creator(extra_environment_variables):
    def create_from_template(file, expect_change=False):
        template_file = file + ".template"
        if not os.path.exists(template_file):
            raise ValueError(f"The template file {template_file} does not exist.")
        warning = (f"The file {file} has unexpectedly changed. You may need to make build-docker-local-clean."
                   if not expect_change
                   else None)
        return create_file_from_template(template_file=template_file,
                                         to_file=file,
                                         extra_environment_variables=extra_environment_variables,
                                         omittable=empty_assignment,
                                         warn_if_changed=warning)
    return create_from_template


def prepare_docker(data_set=DEFAULT_DATA_SET, load_inserts=False, run_tests=False, s3_encrypt_key_id=""):
    extra_vars = {
        "DATA_SET": data_set,
        "LOAD_INSERTS": "true" if load_inserts else "",
        "RUN_TESTS": "true" if run_tests else "",
        "S3_ENCRYPT_KEY_ID": s3_encrypt_key_id or "",
    }
    prepare_from_template = template_creator(extra_vars)
    prepare_from_template(PrepOptions.DOCKER_COMPOSE_FILE, expect_change=True)
    prepare_from_template(PrepOptions.DOCKER_DEVELOPMENT_INI_FILE)


def prepare_docker_main():
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Prepare docker files", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("--data-set", default=DEFAULT_DATA_SET, choices=DATA_SET_CHOICES,
                        help=f"the data set to use (default: {DEFAULT_DATA_SET})")
    parser.add_argument("--load-inserts", default=False, action="store_true",
                        help="if supplied, causes inserts to be loaded (default: not loaded)")
    parser.add_argument("--run-tests", default=False, action="store_true",
                        help="if supplied, causes tests to be run in container (default: not tested)")
    parser.add_argument('--s3-encrypt-key-id', default=None,
                        help="an encrypt key id (default: None)")
    args = parser.parse_args()

    logging.basicConfig()
    prepare_docker(data_set=args.data_set,
                   load_inserts=args.load_inserts,
                   run_tests=args.run_tests,
                   s3_encrypt_key_id=args.s3_encrypt_key_id)


def prepare_local_dev(force=False):
    extra_vars = {}
    prepare_from_template = template_creator(extra_vars)
    for file in [PrepOptions.TEST_INI_FILE, PrepOptions.DEVELOPMENT_INI_FILE]:
        exists = os.path.exists(file)
        if not exists or force:
            reason = "it doesn't exist" if not exists else "--force was given to prepare-local-dev"
            PRINT(f"{'Recreating' if exists else 'Creating'} {file} because {reason} ...")
            prepare_from_template(file, expect_change=True)
            PRINT(f"{'Recreated' if exists else 'Created'} {file}.")
        else:
            PRINT(f"The file {file} already exists and will not be changed.")


def prepare_local_dev_main():
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Prepare files used for local development (test.ini and development.ini)", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument("--force", default=False, action="store_true",
                        help="forces creation of a development.ini and test.ini even if they already exist."
                             " By default, creation is skipped for any file that already exists.")
    args = parser.parse_args()

    logging.basicConfig()
    prepare_local_dev(force=args.force)



from dcicutils.project_utils import C4ProjectRegistry, C4Project
from .project.authentication import SnovaultProjectAuthentication
from .project.authorization import SnovaultProjectAuthorization
from .project.ingestion import SnovaultProjectIngestion


@C4ProjectRegistry.register("dcicsnovault")
class SnovaultProject(SnovaultProjectAuthentication,
                      SnovaultProjectAuthorization,
                      SnovaultProjectIngestion,
                      C4Project):
    NAMES = {"NAME": "snovault", "PYPI_NAME": "dcicsnovault"}
    ACCESSION_PREFIX = "SNO"



import contextlib
import pytest

from dcicutils.lang_utils import disjoined_list
from dcicutils.qa_utils import logged_messages, input_mocked
from unittest import mock
from ..commands import clear_db_es_contents as clear_db_es_contents_module
from ..commands.clear_db_es_contents import (
    clear_db_tables,
    run_clear_db_es,
    main as clear_db_es_contents_main
)


pytestmark = [pytest.mark.setone, pytest.mark.working, pytest.mark.indexing]


def test_clear_db_tables(app, testapp):
    # post an item and make sure it's there
    post_res = testapp.post_json('/testing-post-put-patch-sno/', {'required': 'abc'},
                                 status=201)
    testapp.get(post_res.location, status=200)
    clear_res = clear_db_tables(app)
    assert clear_res is True
    # item should no longer be present
    testapp.get(post_res.location, status=404)


_FOURFRONT_PRODUCTION_ENVS = ['fourfront-production-blue', 'fourfront-production-green', 'data', 'staging']
# Really we only care about the first of these names, but the rest are names that were at one time
# planned to be stg or prd names for cgap, so we'll use them to tell that run_clear_db_es is properly
# skipping any such names. -kmp 4-Jun-2022
_CGAP_PRODUCTION_ENVS = ['fourfront-cgap', 'fourfront-cgap-green', 'cgap-green', 'fourfront-cgap-blue', 'cgap-blue']

_PRODUCTION_ENVS = [_FOURFRONT_PRODUCTION_ENVS] + [_CGAP_PRODUCTION_ENVS]

TEST_ENV = 'cgap-devtest'

OTHER_ENV = 'fourfront-foo'

DECOY_ENV_1 = TEST_ENV + '-decoy-1'
DECOY_ENV_2 = TEST_ENV + '-decoy-2'


@contextlib.contextmanager
def local_env_name_registry_setting_for_testing(app, envname):
    old_env = app.registry.settings.get('env.name')
    print(f"Remembering old env.name = {old_env}")
    try:
        app.registry.settings['env.name'] = envname
        print(f"Set env.name = {envname}")
        yield
    finally:
        if old_env is None:
            print(f"Removing env.name")
            del app.registry.settings['env.name']
        else:
            print(f"Restoring env.name to {old_env}")
            app.registry.settings['env.name'] = old_env


@pytest.mark.unit
def test_run_clear_db_es_unit(app, testapp):

    with mock.patch.object(clear_db_es_contents_module, "clear_db_tables") as mock_clear_db_tables:
        with mock.patch.object(clear_db_es_contents_module, "run_create_mapping") as mock_run_create_mapping:

            def mocked_is_stg_or_prd_env(env):
                result = (env in _PRODUCTION_ENVS  # really this should be enough
                          # for pragmatic redundancy since these will match our real production systems, protect them
                          or env in _CGAP_PRODUCTION_ENVS
                          or env in _FOURFRONT_PRODUCTION_ENVS
                          or env.endswith("blue") or env.endswith("green") or env.endswith("cgap"))
                print(f"Mocked is_stg_or_prd_env({env}) returning {result}.")
                return result

            with mock.patch.object(clear_db_es_contents_module, "is_stg_or_prd_env") as mock_is_stg_or_prd_env:
                mock_is_stg_or_prd_env.side_effect = mocked_is_stg_or_prd_env

                expected_db_clears = 0
                expected_es_clears = 0

                assert mock_clear_db_tables.call_count == expected_db_clears
                assert mock_run_create_mapping.call_count == expected_es_clears

                # It works positionally
                assert run_clear_db_es(app, None, True) is True
                expected_db_clears += 1
                expected_es_clears += 0

                assert mock_clear_db_tables.call_count == expected_db_clears
                assert mock_run_create_mapping.call_count == expected_es_clears

                # It works by keyword argument
                assert run_clear_db_es(app, only_envs=None, skip_es=True) is True
                expected_db_clears += 1
                expected_es_clears += 0

                assert mock_clear_db_tables.call_count == expected_db_clears
                assert mock_run_create_mapping.call_count == expected_es_clears

                for production_env in _PRODUCTION_ENVS:
                    with local_env_name_registry_setting_for_testing(app, production_env):
                        # should never run on production envs env
                        assert clear_db_es_contents_module.is_stg_or_prd_env(production_env) is True
                        with logged_messages(module=clear_db_es_contents_module, error=[
                            (f'clear_db_es_contents: This action cannot be performed on env {production_env}'
                             f' because it is a production-class (stg or prd) environment.'
                             f' Skipping the attempt to clear DB.')]):
                            assert run_clear_db_es(app, only_envs=None, skip_es=True) is False
                        expected_db_clears += 0
                        expected_es_clears += 0
                        assert mock_clear_db_tables.call_count == expected_db_clears
                        assert mock_run_create_mapping.call_count == expected_es_clears

                with local_env_name_registry_setting_for_testing(app, TEST_ENV):

                    allowed_envs = [OTHER_ENV]

                    # test if we are only running on specific envs
                    with logged_messages(module=clear_db_es_contents_module,
                                         error=[(f'clear_db_es_contents: The current environment, {TEST_ENV},'
                                                 f' is not {disjoined_list(allowed_envs)}.'
                                                 f' Skipping the attempt to clear DB.')]):
                        assert run_clear_db_es(app, only_envs=allowed_envs, skip_es=True) is False
                    expected_db_clears += 0
                    expected_es_clears += 0
                    assert mock_clear_db_tables.call_count == expected_db_clears
                    assert mock_run_create_mapping.call_count == expected_es_clears

                    # test again if we are only running on specific envs
                    with logged_messages(module=clear_db_es_contents_module,
                                         error=[(f'clear_db_es_contents: The current environment, {TEST_ENV},'
                                                 f' is not {disjoined_list(allowed_envs)}.'
                                                 f' Skipping the attempt to clear DB.')]):
                        assert run_clear_db_es(app, only_envs=allowed_envs, skip_es=False) is False
                    expected_db_clears += 0
                    expected_es_clears += 0
                    assert mock_clear_db_tables.call_count == expected_db_clears
                    assert mock_run_create_mapping.call_count == expected_es_clears

                    # test if we are only running on specific envs
                    assert run_clear_db_es(app, only_envs=[TEST_ENV], skip_es=True) is True
                    expected_db_clears += 1
                    expected_es_clears += 0
                    assert mock_clear_db_tables.call_count == expected_db_clears
                    assert mock_run_create_mapping.call_count == expected_es_clears

                    # test again if we are only running on specific envs
                    assert run_clear_db_es(app, only_envs=[TEST_ENV], skip_es=False) is True
                    expected_db_clears += 1
                    expected_es_clears += 1
                    assert mock_clear_db_tables.call_count == expected_db_clears
                    assert mock_run_create_mapping.call_count == expected_es_clears

                    allowed_envs = [DECOY_ENV_1, DECOY_ENV_2]
                    # test if we are only running on specific envs
                    with logged_messages(module=clear_db_es_contents_module,
                                         error=[(f'clear_db_es_contents: The current environment, {TEST_ENV},'
                                                 f' is not {disjoined_list(allowed_envs)}.'
                                                 f' Skipping the attempt to clear DB.')]):
                        assert run_clear_db_es(app, only_envs=allowed_envs, skip_es=False) is False
                    expected_db_clears += 0
                    expected_es_clears += 0
                    assert mock_clear_db_tables.call_count == expected_db_clears
                    assert mock_run_create_mapping.call_count == expected_es_clears

                    allowed_envs = [DECOY_ENV_1, TEST_ENV]
                    # test if we are only running on specific envs
                    assert run_clear_db_es(app, only_envs=allowed_envs, skip_es=False) is True
                    expected_db_clears += 1
                    expected_es_clears += 1
                    assert mock_clear_db_tables.call_count == expected_db_clears
                    assert mock_run_create_mapping.call_count == expected_es_clears


@pytest.mark.unit
def test_clear_db_es_contents_main():

    # It should never get to these first two in this test, but they're ethere for safety.
    with mock.patch.object(clear_db_es_contents_module, "clear_db_tables"):
        with mock.patch.object(clear_db_es_contents_module, "run_create_mapping"):

            class FakeApp:

                class Registry:
                    def __init__(self):
                        self.settings = {}

                def __init__(self, config_uri, appname):
                    self.appname = appname
                    self.config_uri = config_uri
                    self.registry = self.Registry()

                def __str__(self):
                    return f"<FakeApp {self.appname} {self.config_uri} {id(self)}>"

                def __repr__(self):
                    return str(self)

            class MockDBSession:

                def __init__(self, app):
                    self.app = app

            apps = {}

            def mocked_get_app(config_uri, appname):
                key = (config_uri, appname)
                app = apps.get(key)
                if not app:
                    apps[key] = app = FakeApp(config_uri, appname)
                return app

            def mocked_configure_dbsession(app):
                return MockDBSession(app)

            with mock.patch.object(clear_db_es_contents_module, "run_clear_db_es") as mock_run_clear_db_es:
                with mock.patch.object(clear_db_es_contents_module, "get_app") as mock_get_app:
                    mock_get_app.side_effect = mocked_get_app
                    with mock.patch.object(clear_db_es_contents_module,
                                           "configure_dbsession") as mock_configure_dbsession:
                        mock_configure_dbsession.side_effect = mocked_configure_dbsession

                        config_uri = 'production.ini'
                        appname = "app"


                        with input_mocked(
                                # We'll be prompted for the environment name to confirm.
                                "local",
                                module=clear_db_es_contents_module):

                            clear_db_es_contents_main([config_uri])
                            mock_run_clear_db_es.assert_called_with(app=mocked_get_app(config_uri, None),
                                                                    only_envs=[],
                                                                    skip_es=False,
                                                                    allow_prod=False)

                        with input_mocked(
                                # No input prompting will occur because --no-confirm was supplied.
                                module=clear_db_es_contents_module):

                            clear_db_es_contents_main([config_uri, "--no-confirm"])
                            mock_run_clear_db_es.assert_called_with(app=mocked_get_app(config_uri, None),
                                                                    only_envs=[],
                                                                    skip_es=False,
                                                                    allow_prod=False)

                        with input_mocked(
                                # We'll be prompted for the environment name to confirm.
                                "local",
                                module=clear_db_es_contents_module):

                            clear_db_es_contents_main([config_uri, "--app-name", appname])
                            mock_run_clear_db_es.assert_called_with(app=mocked_get_app(config_uri, appname),
                                                                    only_envs=[],
                                                                    skip_es=False,
                                                                    allow_prod=False)

                        with input_mocked(
                                # We'll be prompted for the environment name to confirm.
                                "local",
                                module=clear_db_es_contents_module):

                            clear_db_es_contents_main([config_uri, "--app-name", appname, '--skip-es'])
                            mock_run_clear_db_es.assert_called_with(app=mocked_get_app(config_uri, appname),
                                                                    only_envs=[],
                                                                    skip_es=True,
                                                                    allow_prod=False)

                        with input_mocked(
                                # No input prompting will occur because --only-if-env was supplied.
                                module=clear_db_es_contents_module):

                            clear_db_es_contents_main([config_uri, "--app-name", appname, "--only-if-env", TEST_ENV])
                            mock_run_clear_db_es.assert_called_with(app=mocked_get_app(config_uri, appname),
                                                                    only_envs=[TEST_ENV],
                                                                    skip_es=False,
                                                                    allow_prod=False)

                        with input_mocked(
                                # We'll be prompted for the environment name to confirm.
                                "local",
                                module=clear_db_es_contents_module):

                            clear_db_es_contents_main([config_uri, "--app-name", appname, "--only-if-env", TEST_ENV,
                                                       "--confirm"])
                            mock_run_clear_db_es.assert_called_with(app=mocked_get_app(config_uri, appname),
                                                                    only_envs=[TEST_ENV],
                                                                    skip_es=False,
                                                                    allow_prod=False)

                        with input_mocked(
                                # No input prompting will occur because --only-if-env was supplied.
                                module=clear_db_es_contents_module):

                            clear_db_es_contents_main([config_uri, "--app-name", appname,
                                                       "--only-if-env", f"{TEST_ENV},{OTHER_ENV}"])
                            mock_run_clear_db_es.assert_called_with(app=mocked_get_app(config_uri, appname),
                                                                    only_envs=[TEST_ENV, OTHER_ENV],
                                                                    skip_es=False,
                                                                    allow_prod=False)
